use warnings;
use strict;
use Test::More tests => 5;
use Test::Exception;
use Test::MockModule;

use POWER::Event;
use POWER::Event::IO;


POWER::Event::IO->new({
    type        => $LISTEN,
    host        => '0.0.0.0',
    port        => 1234,
    cb          => \&listener,
    wait_for    => $ACCEPTED,
});
POWER::Event::IO->new({
    host        => '127.0.0.1',
    port        => 1234,
    cb          => \&client,
    wait_for    => $EOF,
    out_buf     => 'test',
    in_buf_limit=> 1024,
});

my @Wait = (
    ['listener',    $ACCEPTED   ],
    ['server',      $IN         ],
    ['client',      $EOF        ],
);

eval { loop() };
die "out of loop: $@";

sub listener {
    my ($io) = @_;
    is_deeply(shift @Wait, ['listener', $io->{events}], 'listener');
    die "listener unexpected event: $io->{err}" if $io->{events} != $ACCEPTED;
    POWER::Event::IO->new({
        fh          => delete($io->{in_buf}),
        cb          => \&server,
        wait_for    => $IN,
    });
}

sub server {
    my ($io) = @_;
    is_deeply(shift @Wait, ['server', $io->{events}], 'server');
    die "server unexpected event: $io->{err}" if $io->{events} != $IN;
    is($io->{in_buf}, 'test', '  got "test" request');
    $io->write('test passed');
    shutdown $io->{fh}, 1;
}

sub client {
    my ($io) = @_;
    is_deeply(shift @Wait, ['client', $io->{events}], 'client');
    die "server unexpected event: $io->{err}" if $io->{events} != $EOF;
    is($io->{in_buf}, 'test passed', '  got "test passed" reply');
    exit;
}

