use warnings;
use strict;
use Test::More tests => 44;
use Test::Exception;
use Test::MockModule;

use Time::HiRes qw( time sleep );

use POWER::Epoll;
use POWER::Event;
use POWER::Event::Timer;
use POWER::Event::IO;


my ($DNS_RESEND, @Event, $io, $timer, $module, $module2) = (0);

#      fd   $DNS
my ($dns) = `ls -al /proc/$$/fd/ | grep socket | head -n 1`=~/(\d+) -> socket/;


if ($INC{'Devel/Cover.pm'}) {
$POWER::Event::IO::TO_DNS       = int 7 / $POWER::Event::IO::TICK_RATE + 0.5;
$POWER::Event::IO::TO_DNS_RESEND= int 2 / $POWER::Event::IO::TICK_RATE + 0.5;
$POWER::Event::IO::TO_CONNECT   = int 3 / $POWER::Event::IO::TICK_RATE + 0.5;
$POWER::Event::IO::TO_WRITE     = int 3 / $POWER::Event::IO::TICK_RATE + 0.5;
} else {
$POWER::Event::IO::TO_DNS       = int 0.7 / $POWER::Event::IO::TICK_RATE + 0.5;
$POWER::Event::IO::TO_DNS_RESEND= int 0.2 / $POWER::Event::IO::TICK_RATE + 0.5;
$POWER::Event::IO::TO_CONNECT   = int 0.3 / $POWER::Event::IO::TICK_RATE + 0.5;
$POWER::Event::IO::TO_WRITE     = int 0.3 / $POWER::Event::IO::TICK_RATE + 0.5;
}


sub _loop {
    eval { loop() }; die "died in loop: $@" if $@ !~ /timer/;
}

sub cb {
    my ($io) = @_;
    push @Event, [ $io->{events}, $io->{err} ];
}

if ($INC{'Devel/Cover.pm'}) {
$timer = set_timer({ every => 1, cb => sub { die "timer\n" } });
} else {
$timer = set_timer({ every => 0.1, cb => sub { die "timer\n" } });
}

#diag ">>>>>                 <<<<<";
#diag ">>>>> testing TO_DNS* <<<<<";
#diag ">>>>>                 <<<<<";

lives_ok { $io=POWER::Event::IO->new({cb=>\&cb,host=>'localhost',port=>80,
                                      wait_for=>$RESOLVED|$CONNECTED}) };
$module  = Test::MockModule->new('POWER::Event::IO');
$module->mock('epoll_wait', sub { sleep($_[1]); return(0); });
$module2 = Test::MockModule->new('POWER::Event::IO::UDP_CONNECTED', no_auto=>1);
$module2->mock('_real_write', sub {
    $DNS_RESEND++;
    delete $_[0]->{out_buf};
    return 1;
});
_loop();
is(@Event, 0, 'no events after 0.00-0.09 sec');
_loop();
is(@Event, 0, 'no events after 0.10-0.19 sec');
is($DNS_RESEND, 0, 'no DNS packet was resent yet');
_loop();
is(@Event, 0, 'no events after 0.20-0.29 sec');
is($DNS_RESEND, 1, 'DNS packet sent again');
_loop();
is(@Event, 0, 'no events after 0.30-0.39 sec');
is($DNS_RESEND, 1, '');
_loop();
is(@Event, 0, 'no events after 0.40-0.49 sec');
is($DNS_RESEND, 2, 'DNS packet sent again');
_loop();
is(@Event, 0, 'no events after 0.50-0.59 sec');
is($DNS_RESEND, 2, '');
_loop();
is(@Event, 0, 'no events after 0.60-0.69 sec');
is($DNS_RESEND, 3, 'DNS packet sent again');
_loop();
is(@Event, 1, 'got 1 event after 0.70-0.79 sec');
is_deeply(shift @Event, [$ERR, q{resolving timed out}], 'TO_DNS');
_loop();
is(@Event, 0, 'no events after 0.80-0.89 sec');
_loop();
is(@Event, 0, 'no events after 0.90-0.99 sec');
is($DNS_RESEND, 3, 'no DNS packets was sent after TO_DNS');
$module->unmock_all();
$module2->unmock_all();
$io->close();


#diag ">>>>>                    <<<<<";
#diag ">>>>> testing TO_CONNECT <<<<<";
#diag ">>>>>                    <<<<<";

lives_ok { $io=POWER::Event::IO->new({cb=>\&cb,host=>'localhost',port=>80,
                                      wait_for=>$RESOLVED|$CONNECTED}) };
$module = Test::MockModule->new('POWER::Event::IO');
$module->mock('epoll_wait', sub { return (0); });
_loop();
is(@Event, 0, 'no events after 0.00-0.09 sec');
$module->mock('epoll_wait', sub {
    my ($status, $events) = $module->original('epoll_wait')->(@_);
    if ($status) {
        $events = [ grep { $_->[0] == $dns } @$events ];
        $status = @$events;
    }
    return ($status, $events);
});
_loop();
is(@Event, 1, 'got 1 event after 0.10-0.19 sec');
is_deeply(shift @Event, [$RESOLVED, undef], 'RESOLVED');
_loop();
is(@Event, 0, 'no events after 0.20-0.29 sec');
_loop();
is(@Event, 0, 'no events after 0.30-0.39 sec');
_loop();
is(@Event, 1, 'got 1 event after 0.40-0.49 sec');
is_deeply(shift @Event, [$ERR, 'connecting timed out'], 'TO_CONNECT');
_loop();
is(@Event, 0, 'no events after 0.50-0.59 sec');
_loop();
is(@Event, 0, 'no events after 0.60-0.69 sec');
_loop();
is(@Event, 0, 'no events after 0.70-0.79 sec');
_loop();
is(@Event, 0, 'no events after 0.80-0.89 sec');
$module->unmock_all();
$io->close();


#diag ">>>>>                  <<<<<";
#diag ">>>>> testing TO_WRITE <<<<<";
#diag ">>>>>                  <<<<<";

sub li { POWER::Event::IO->new({cb => sub{}, fh => delete $_[0]->{in_buf} }) }
lives_ok {     POWER::Event::IO->new({cb=>\&li,host=>'localhost',port=>1234,
                                      type=>$LISTEN,wait_for=>$ACCEPTED}) };
lives_ok { $io=POWER::Event::IO->new({cb=>\&cb,host=>'localhost',port=>1234,
                                      wait_for=>$RESOLVED|$CONNECTED}) };
$module = Test::MockModule->new('POWER::Event::IO');
_loop();
is(@Event, 1, 'got 1 event after 0.00-0.09 sec');
is_deeply(shift @Event, [$RESOLVED|$CONNECTED, undef], 'RESOLVED&CONNECTED');
$module->mock('epoll_wait', sub {
    my ($status, $events) = $module->original('epoll_wait')->(@_);
    if ($status) {
        $events = [ grep { $_->[0] != $io->{fd} } @$events ];
        $status = @$events;
    }
    return ($status, $events);
});
$module2 = Test::MockModule->new('POWER::Event::IO::STREAM', no_auto=>1);
$module2->mock('_real_write', sub {
    return 0;
});
$io->write("test");
_loop();
is(@Event, 0, 'no events after 0.10-0.19 sec');
_loop();
is(@Event, 0, 'no events after 0.20-0.29 sec');
_loop();
is(@Event, 0, 'no events after 0.30-0.39 sec');
_loop();
is(@Event, 1, 'got 1 event after 0.40-0.49 sec');
is_deeply(shift @Event, [$ERR, 'writing timed out'], 'TO_WRITE');
_loop();
is(@Event, 0, 'no events after 0.50-0.59 sec');
_loop();
is(@Event, 0, 'no events after 0.60-0.69 sec');
_loop();
is(@Event, 0, 'no events after 0.70-0.79 sec');
_loop();
is(@Event, 0, 'no events after 0.80-0.89 sec');
$module->unmock_all();
$module2->unmock_all();
$io->close();


